from networkx.linalg.attrmatrix import *
from networkx.linalg import attrmatrix
from networkx.linalg.spectrum import *
from networkx.linalg import spectrum
from networkx.linalg.graphmatrix import *
from networkx.linalg import graphmatrix
from networkx.linalg.laplacianmatrix import *
from networkx.linalg import laplacianmatrix
from networkx.linalg.algebraicconnectivity import *
from networkx.linalg.modularitymatrix import *
from networkx.linalg import modularitymatrix
from networkx.linalg.bethehessianmatrix import *
from networkx.linalg import bethehessianmatrix
